/* lunarlander.c - lunar lander game
 *
 * Copyright 2003 Ben Tayler

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#define SCREEN_WIDTH 640
#define SCREEN_HEIGHT 448
#define SCREEN_DEPTH 32
#define MAX_FPS 50

#define BG_FILENAME "background.bmp"
#define LANDER_FILENAME "lander_.bmp"
#define LANDER_L_FILENAME "lander_l.bmp"
#define LANDER_R_FILENAME "lander_r.bmp"
#define LANDER_LR_FILENAME "lander_lr.bmp"
#define BLANK_FILENAME "blank.bmp"
#define METEOR_FILENAME "meteor.bmp"
#define PAD_FILENAME "pad_.bmp"
#define PAD_L_FILENAME "pad_l.bmp"
#define SHADOW_FILENAME "shadow.bmp"
#define FIRE_FILENAME "fire.bmp"
#define FUELBAR_FILENAME "fuelbar.bmp"
#define SPEED_FILENAME "speed.bmp"
#define NUMBERS_FILENAME "numbers.bmp"
#define LIFEICON_FILENAME "tinylander.bmp"
#define BIGBLANK_FILENAME "bigblank.bmp"
#define UFO_FILENAME "ufo.bmp"

#define METEOR_ALPHA 64
#define NUM_METEORS 8
#define MAX_METEOR_SPEED 1.5f
#define MIN_METEOR_SPEED 0.5f
#define MAX_SHADOW_ALPHA 192
#define FIRE_ALPHA 128
#define LANDER_VBORDER 18
#define LANDER_XBORDER 11
#define PAD_BORDER 5

#define FUELBAR_X 590
#define FUELBAR_Y 40
#define FUEL_WARNING 0.25f
#define FLASH_INTERVAL 4.0f
#define FUELBAR_ALPHA 128
#define SPEEDTEXT_X 50
#define SPEEDTEXT_Y 40
#define NUMBER_SPACING 17
#define HSPEED_X 75
#define HSPEED_Y 46
#define VSPEED_X 81
#define VSPEED_Y 82
#define LIFEICON_X 50
#define LIFEICON_Y 113
#define LIFEICON_SPACING 6
#define FADE_FRAMES 10
#define FADE_ALPHA 128

#define UFO_Y 100
#define MAX_UFO_DELAY 5.0f
#define UFO_SPEED 3.0f

#define MAX_X 110.0f
#define MAX_Y 70.0f
#define X_START 105.0f
#define Y_START 70.0f
#define VX_START -5.0f
#define VY_START 0.0f
#define GRAVITY -1.62f
#define MAX_THRUST 44000.0f
#define EXHAUST_V 4500.0f
#define START_MASS 14700.0f
#define FUEL_MASS 10500.0f
#define ROTATE_PERIOD 4.0f
#define GROUND_LEVEL 40
#define LUNAR_SURFACE 370
#define LANDER_RADIUS 4.0f
#define CRASH_SLOWDOWN 0.5f
#define RESTART_DELAY 5.0f
#define START_LIVES 3

#define SAFE_VX 1.0f
#define SAFE_VY 1.0f
#define SAFE_ANGLE 1.0f

#define USE_PAD 0
#define THRUST_BUTTON 1
#define PAUSE_BUTTON 9
#define QUIT_BUTTON1 8
#define QUIT_BUTTON2 9
#define QUIT_KEY SDLK_ESCAPE
#define ROTATION_AXIS 0
#define DEAD_ZONE 10000

#define MIX_FREQ 22050
#define BGMUSIC_FILENAME "music.mod"
#define MIX_CHANNELS 2

#include <SDL.h>
#include <SDL_mixer.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>

SDL_Surface *framebuf; /* globally accessible framebuffer surface */

void InitSDL()
{
  printf("Initializing SDL.\n");

  /* init sdl modules */
  if((SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK | SDL_INIT_TIMER | SDL_INIT_AUDIO) == -1)) {
    printf("Could not initialize SDL: %s.\n", SDL_GetError());
    exit(-1);
  }

  printf("SDL initialized.\n");
}

void InitMixer()
{
  /* open audio mixer */
  if(Mix_OpenAudio(MIX_FREQ, MIX_DEFAULT_FORMAT, MIX_CHANNELS, 1024) == -1) {
    printf("Mix_OpenAudio: %s\n", Mix_GetError());
    exit(2);
  }

}

void QuitSDL()
{
 printf("Quitting SDL.\n");

  /* shutdown all subsystems */
  SDL_Quit();

  printf("Quitting...\n");

  exit(0);
}

void SetupScreen()
{
  /* request a double-buffered hardware surface*/
  framebuf = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_DEPTH, SDL_HWSURFACE 
			      | SDL_DOUBLEBUF);

  if ( framebuf == NULL ) {
    fprintf(stderr, "Couldn't set %dx%dx%d video mode: %s\n", SCREEN_WIDTH,
	    SCREEN_HEIGHT, SCREEN_DEPTH, SDL_GetError());
    exit(1);
  }
}

SDL_Rect BlitSurface(SDL_Surface *image, SDL_Surface *screen, int x, int y)
{
    SDL_Rect dest;

    /* perform blit with surfaces unlocked */
    dest.x = x;
    dest.y = y;
    dest.w = image->w;
    dest.h = image->h;

    SDL_BlitSurface(image, NULL, screen, &dest);

    return dest;
}

Uint32 GetPixel(SDL_Surface *surface, int x, int y)
{
  int bpp = surface->format->BytesPerPixel;

  /* pixel address p */
  Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

  return p[0] | p[1] << 8 | p[2] << 16;

}

void PutPixel(SDL_Surface *surface, int x, int y, Uint32 pixel)
{
  int bpp = surface->format->BytesPerPixel;

  /* pixel address p */
  Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

  p[0] = pixel & 0xff;
  p[1] = (pixel >> 8) & 0xff;
  p[2] = (pixel >> 16) & 0xff;

}

void LockSurface(SDL_Surface *surface)
{
  if (SDL_MUSTLOCK(surface))
    SDL_LockSurface(surface);
}

void UnlockSurface(SDL_Surface *surface)
{
  if (SDL_MUSTLOCK(surface))
    SDL_UnlockSurface(surface);
}

void RotateSurface(SDL_Surface *source, SDL_Surface *dest, int originx, int originy, float angle)
{
  float sintemp, costemp;
  int xloop, yloop, rotatedx, rotatedy;
  Uint32 pixel;

  LockSurface(source);
  LockSurface(dest);

  /* precalc trig functions */
  sintemp = sinf(angle);
  costemp = cosf(angle);

  for (xloop = 0; xloop < source->w; ++xloop)
    for (yloop = 0; yloop < source->h; ++yloop) {

      rotatedx = (-1 * (((originx - xloop) * costemp) - ((originy - yloop) * sintemp))) 
	+ originx + ((dest->w - source->w) / 2);
      rotatedy = (-1 * (((originx - xloop) * sintemp) + ((originy - yloop) * costemp))) 
	+ originy + ((dest->h - source->h) / 2);

      /* cover holes in rotated sprite */
      pixel = GetPixel(source, xloop, yloop);

      PutPixel(dest, rotatedx, rotatedy, pixel);
      PutPixel(dest, rotatedx + 1, rotatedy, pixel);
      PutPixel(dest, rotatedx, rotatedy + 1, pixel);
    }

  UnlockSurface(source);
  UnlockSurface(dest);
}

float GetJoyScale(SDL_Joystick *controller, int axis)
{
  int joypos;
  float scalefactor;

  if (abs(joypos = SDL_JoystickGetAxis(controller, axis)) > DEAD_ZONE)
    scalefactor = joypos / 32768.0f;
  else
    scalefactor = 0.0f;

  return scalefactor;
}

int CheckForPause(SDL_Joystick *controller)
{
  int pause;

  if (SDL_JoystickGetButton(controller, PAUSE_BUTTON)) {

    pause = 1;

    while (SDL_JoystickGetButton(controller, PAUSE_BUTTON))
      SDL_PumpEvents();
    while ( ! SDL_JoystickGetButton(controller, PAUSE_BUTTON))
      SDL_PumpEvents();
    while (SDL_JoystickGetButton(controller, PAUSE_BUTTON))
      SDL_PumpEvents();

  }
  else pause = 0;

  return pause;
}

int CheckLanding(float vx, float vy, float angle)
{
  int goodlanding;

  if (fabs(vx) < SAFE_VX && fabs(vy) < SAFE_VY && fabs(angle) < SAFE_ANGLE)
    goodlanding = 1;
  else
    goodlanding = 0;

  return goodlanding;
}

void BlitNumber(SDL_Surface *source, SDL_Surface *dest, float number, int pixelx, int pixely)
{
  SDL_Rect cliparea;

  /* set clip area for first digit */
  cliparea.x = pixelx;
  cliparea.y = pixely;
  cliparea.w = NUMBER_SPACING;
  cliparea.h = source->h;

  number = fabs(number);

  if (number >= 100.0f)
    number = 99.0f;

  /* tens */
  if (number >= 10.0f) {
    SDL_SetClipRect(dest, &cliparea);
    BlitSurface(source, dest, (pixelx - (floor(number / 10) * NUMBER_SPACING)), pixely);

    /* set clip area for next digit */
    pixelx += NUMBER_SPACING;
    cliparea.x = pixelx;
  }

  /* units */
  SDL_SetClipRect(dest, &cliparea);
  BlitSurface(source, dest, (pixelx - (floor(number - (floor(number / 10) * 10)) 
				       * NUMBER_SPACING)), pixely);
  pixelx += NUMBER_SPACING;
  cliparea.x = pixelx;

  /* fractional part */
  if (number <= 10.0f) {

    /* blit decimal point */
    SDL_SetClipRect(dest, &cliparea);
    BlitSurface(source, dest, (pixelx - (10 * NUMBER_SPACING)), pixely);

    pixelx += source->w - (10 * NUMBER_SPACING);
    cliparea.x = pixelx;

    /* blit fraction */
    SDL_SetClipRect(dest, &cliparea);
    BlitSurface(source, dest, (pixelx - (floor(fmod(number, 1.0f) * 10) * NUMBER_SPACING)), pixely);
  }
}

int main()
{
  /* set up surfaces */
  SDL_Surface *background;

  SDL_Surface *lander_;
  SDL_Surface *lander_l;
  SDL_Surface *lander_r;
  SDL_Surface *lander_lr;
  SDL_Surface *spriteptr;
  SDL_Surface *landerfinal;

  SDL_Surface *meteor;
  SDL_Surface *pad_;
  SDL_Surface *pad_l;
  SDL_Surface *padptr;
  SDL_Surface *shadow;
  SDL_Surface *fire;

  SDL_Surface *fuelbar;
  SDL_Surface *speedtext;
  SDL_Surface *numbers;
  SDL_Surface *lifeicon;
  SDL_Surface *fadesurface;

  SDL_Surface *ufo;

  /* clipping areas */
  SDL_Rect cliparea, meteorarea[NUM_METEORS], padarea, shadowarea, lifeiconarea, ufoarea;
  SDL_Rect fuelbararea, speedarea;

  /* input variables */
  SDL_Joystick *controller;

  /* audio chunks */
  Mix_Music *bgmusic;

  /* set up keypress array */
  Uint8 *kbstate = SDL_GetKeyState(NULL);

  /* simulation variables */
  float x, y, dx, dy, vx, vy, dvx, dvy, ax, ay, mass, dt;
  float angle, dangle;
  float crashvx, crashvy;

  int pixelx, pixely, oldpixelx, oldpixely;

  int onground, crashed, gotcrashinfo;
  int restartflag;
  int livesleft;

  /* effects variables */
  int meteorx[NUM_METEORS], meteory[NUM_METEORS], oldmeteorx[NUM_METEORS], oldmeteory[NUM_METEORS];
  float meteorspeed[NUM_METEORS];
  int firealpha;
  int loop;
  int displayfuel, displayspeed;

  int ufopixelx, ufopixely, oldufopixelx, oldufopixely, ufoframes, ufodirection;

  /* rotation variables */
  int originx, originy;

  /* timing variables */
  int frames;
  float fps;
  Uint32 frametimer, temptimer;
  int meteorframes[NUM_METEORS];
  int fuelbarframes, speedframes, crashframes;

  InitSDL();
  SetupScreen();
  InitMixer();

  /* open controller 1 */
  controller = SDL_JoystickOpen(USE_PAD);
  SDL_JoystickEventState(SDL_ENABLE);

  /* preload surfaces with bitmaps */
  background = SDL_LoadBMP(BG_FILENAME);

  lander_ = SDL_LoadBMP(LANDER_FILENAME);
  lander_l = SDL_LoadBMP(LANDER_L_FILENAME);
  lander_r = SDL_LoadBMP(LANDER_R_FILENAME);
  lander_lr = SDL_LoadBMP(LANDER_LR_FILENAME);
  landerfinal = SDL_LoadBMP(BLANK_FILENAME);

  shadow = SDL_LoadBMP(SHADOW_FILENAME);
  meteor = SDL_LoadBMP(METEOR_FILENAME);
  pad_ = SDL_LoadBMP(PAD_FILENAME);
  pad_l = SDL_LoadBMP(PAD_L_FILENAME);
  fire = SDL_LoadBMP(FIRE_FILENAME);

  fuelbar = SDL_LoadBMP(FUELBAR_FILENAME);
  speedtext = SDL_LoadBMP(SPEED_FILENAME);
  numbers = SDL_LoadBMP(NUMBERS_FILENAME);
  lifeicon = SDL_LoadBMP(LIFEICON_FILENAME);
  fadesurface = SDL_LoadBMP(BIGBLANK_FILENAME);

  ufo = SDL_LoadBMP(UFO_FILENAME);

  if (background == NULL || lander_ == NULL || lander_l == NULL || 
      lander_r == NULL || lander_lr == NULL ||landerfinal == NULL || 
      meteor == NULL || pad_  == NULL || pad_l == NULL || shadow == NULL ||
      fire == NULL || fuelbar == NULL || speedtext == NULL || 
      numbers == NULL || lifeicon == NULL || fadesurface == NULL ||
      ufo == NULL) {
      printf("Image load error.\n");
      exit(1);
  }

  /* play background music */
  bgmusic = Mix_LoadMUS(BGMUSIC_FILENAME);

  if(bgmusic == NULL) {
    fprintf(stderr, "Music load error.\n");
  }
  else
    Mix_PlayMusic(bgmusic, -1);

  /* set surface transparencies */
  SDL_SetColorKey(landerfinal, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(landerfinal->format, 0, 0, 0));

  SDL_SetColorKey(pad_, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(pad_->format, 0, 0, 0));
  SDL_SetColorKey(pad_l, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(pad_l->format, 0, 0, 0));
  SDL_SetColorKey(shadow, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(shadow->format, 0, 0, 0));

  SDL_SetColorKey(meteor, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(meteor->format, 0, 0, 0));
  SDL_SetAlpha(meteor, SDL_SRCALPHA | SDL_RLEACCEL, METEOR_ALPHA);

  SDL_SetColorKey(fire, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(fire->format, 0, 0, 0));

  SDL_SetAlpha(fuelbar, SDL_SRCALPHA | SDL_RLEACCEL, FUELBAR_ALPHA);
  SDL_SetColorKey(speedtext, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(speedtext->format, 0, 0, 0));
  SDL_SetColorKey(numbers, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(numbers->format, 0, 0, 0));
  SDL_SetColorKey(lifeicon, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(lifeicon->format, 0, 0, 0));

  SDL_SetColorKey(ufo, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(ufo->format, 0, 0, 0));

  /* initialize random number generator */
  srand(time(NULL));

  /* precalc shadow vertical clip area */
  shadowarea.y = (SCREEN_HEIGHT - 1 - GROUND_LEVEL) - (shadow->h / 2);
  shadowarea.h = shadow->h;

  /* randomize meteor info */
  for (loop = 0; loop < NUM_METEORS; ++loop) {
    meteorx[loop] = 0;
    meteory[loop] = SCREEN_HEIGHT;

    oldmeteorx[loop] = meteorx[loop];
    oldmeteory[loop] = meteory[loop];

    meteorspeed[loop] = MIN_METEOR_SPEED + (((float)rand() / (float)RAND_MAX) * MAX_METEOR_SPEED);
    meteorframes[loop] = ((float)rand() / (float)RAND_MAX) * MAX_FPS;
  }

  /* set initial UFO position */
  ufopixelx = SCREEN_WIDTH;
  ufopixely = UFO_Y;
  oldufopixelx = 0;
  oldufopixely = ufopixely;
  ufoframes = 0;
  ufodirection = 1;

  /* set UI clipping areas */
  fuelbararea.x = FUELBAR_X;
  fuelbararea.y = FUELBAR_Y;
  fuelbararea.w = fuelbar->w;
  fuelbararea.h = fuelbar->h;

  speedarea.x = SPEEDTEXT_X;
  speedarea.y = SPEEDTEXT_Y;
  speedarea.w = speedtext->w;
  speedarea.h = speedtext->h;

  /* set up static variables */
  livesleft = START_LIVES;
  oldpixelx = oldpixely = 0;

  /* set background */
  BlitSurface(background, framebuf, 0, 0);
  SDL_Flip(framebuf);

  while (livesleft >= 0) {

    /* initial simulation parameters*/
    ax = ay = 0.0f;

    x = X_START;
    y = Y_START;
    vx = VX_START;
    vy = VY_START;
    mass = START_MASS;

    angle = 0.0f;

    frames = 0;
    onground = crashed = gotcrashinfo = 0;
    displayfuel = displayspeed = 1;
    fuelbarframes = speedframes = crashframes = 0;
    restartflag = 0;
    firealpha = 0;

    /* set sprites for first frame */
    spriteptr = lander_;
    padptr = pad_l;

    /* precalc centre of rotation */
    originx = spriteptr->w / 2;
    originy = spriteptr->h / 2;

    /* clean up pad */
    if (livesleft < START_LIVES) {
      SDL_SetClipRect(framebuf, &padarea);
      BlitSurface(background, framebuf, 0, 0);
    }

    /* randomize pad location */
    padarea.x = (padptr->w / 2) + ((SCREEN_WIDTH - (2 * padptr->w)) 
				   * ((float)rand() / (float)RAND_MAX));
    padarea.y = (SCREEN_HEIGHT - 1 - GROUND_LEVEL) - (padptr->h / 2);
    padarea.w = padptr->w;
    padarea.h = padptr->h;

    /* clean up lifeicon area */
    if (livesleft < START_LIVES) {
      SDL_SetClipRect(framebuf, &lifeiconarea);
      BlitSurface(background, framebuf, 0, 0);
    }

    /* calculate lifeicon clip area */
    lifeiconarea.x = LIFEICON_X;
    lifeiconarea.y = LIFEICON_Y;

    lifeiconarea.w = ((lifeicon->w + LIFEICON_SPACING) * livesleft) - LIFEICON_SPACING;
    lifeiconarea.h = lifeicon->h;

    /* init frame timer */
    frametimer = SDL_GetTicks();

    /* main loop */
    while (( ! SDL_JoystickGetButton(controller, QUIT_BUTTON1) || 
	     ! SDL_JoystickGetButton(controller, QUIT_BUTTON2)) && 
	   ! kbstate[QUIT_KEY] && ! restartflag) {

      /* create rotated lander surface */
      SDL_FillRect(landerfinal, NULL, SDL_MapRGB(landerfinal->format,0, 0, 0));
      RotateSurface(spriteptr, landerfinal, originx, originy, angle);

      /* draw current position */
      pixelx = (x * ((SCREEN_WIDTH + spriteptr->w) / MAX_X)) - spriteptr->w;
      pixely = SCREEN_HEIGHT  - GROUND_LEVEL + LANDER_VBORDER 
	- (landerfinal->h + (y * ((SCREEN_HEIGHT - GROUND_LEVEL) / MAX_Y)));

      /* calculate lander clip area */
      if (pixelx > oldpixelx) {
	cliparea.x = oldpixelx;
	cliparea.w = pixelx - oldpixelx + landerfinal->w;
      }
      else {
	cliparea.x = pixelx;
	cliparea.w = oldpixelx - pixelx + landerfinal->w;
      }

      if (pixely > oldpixely) {
	cliparea.y = oldpixely;
	cliparea.h = pixely - oldpixely + landerfinal->h;
      }
      else {
	cliparea.y = pixely;
	cliparea.h = oldpixely - pixely + landerfinal->h;
      }

      /* calculate horizontal shadow clip area */
      if (pixelx > oldpixelx) {
	shadowarea.x = oldpixelx;
	shadowarea.w = pixelx - oldpixelx + shadow->w;
      }
      else {
	shadowarea.x = pixelx;
	shadowarea.w = oldpixelx - pixelx + shadow->w;
      }

      /* calculate meteor clip areas */
      for (loop = 0; loop < NUM_METEORS; ++loop) {

	meteorarea[loop].x = meteorx[loop];
	meteorarea[loop].w = oldmeteorx[loop] - meteorx[loop] + meteor->w;

	meteorarea[loop].y = oldmeteory[loop];

	if (meteory[loop] + meteor->h > LUNAR_SURFACE)

	  if (oldmeteory[loop] > LUNAR_SURFACE)
	    meteorarea[loop].h = 0;
	  else
	    meteorarea[loop].h = (meteory[loop] - oldmeteory[loop] + meteor->h) - 
	      ((meteory[loop] + meteor->h) - LUNAR_SURFACE);

	else
	  meteorarea[loop].h = meteory[loop] - oldmeteory[loop] + meteor->h;
      }

      /* calculate ufo clip area */
      if (ufopixelx > oldufopixelx) {
	ufoarea.x = oldufopixelx;
	ufoarea.w = ufopixelx - oldufopixelx + ufo->w;
      }
      else {
	ufoarea.x = ufopixelx;
	ufoarea.w = oldufopixelx - ufopixelx + ufo->w;
      }

      ufoarea.y = UFO_Y;
      ufoarea.h = ufo->h;

      /* clean up fuelbar */
      SDL_SetClipRect(framebuf, &fuelbararea);
      BlitSurface(background, framebuf, 0, 0);

      /* clean up ufo */
      SDL_SetClipRect(framebuf, &ufoarea);
      BlitSurface(background, framebuf, 0, 0);

      /* calculate fuelbar clip area */
      fuelbararea.x = FUELBAR_X;
      fuelbararea.w = fuelbar->w;

      fuelbararea.y = FUELBAR_Y + fuelbar->h - (((mass - START_MASS + FUEL_MASS) 
						 / FUEL_MASS) * fuelbar->h);
      fuelbararea.h = fuelbar->h - fuelbararea.y + FUELBAR_Y;

      /* clean up lander */
      SDL_SetClipRect(framebuf, &cliparea);
      BlitSurface(background, framebuf, 0, 0);

      /* clean up shadow */
      SDL_SetClipRect(framebuf, &shadowarea);
      BlitSurface(background, framebuf, 0, 0);

      /* clean up UI text */
      SDL_SetClipRect(framebuf, &speedarea);
      BlitSurface(background, framebuf, 0, 0);

      /* clean up meteors */
      for (loop = 0; loop < NUM_METEORS; ++loop) {

	SDL_SetClipRect(framebuf, &meteorarea[loop]);
	BlitSurface(background, framebuf, 0, 0);
      }

      /* blit meteors */
      for (loop = 0; loop < NUM_METEORS; ++loop) {

	SDL_SetClipRect(framebuf, &meteorarea[loop]);
	BlitSurface(meteor, framebuf, meteorx[loop], meteory[loop]);
      }

      /* blit UI text */
      SDL_SetClipRect(framebuf, &speedarea);
      BlitSurface(speedtext, framebuf, SPEEDTEXT_X, SPEEDTEXT_Y);

      if (! crashed) {
	BlitNumber(numbers, framebuf, vx, HSPEED_X, HSPEED_Y);
	BlitNumber(numbers, framebuf, vy, VSPEED_X, VSPEED_Y);
      }
      else {

	if (displayspeed || fabs(crashvx) < SAFE_VX)
	  BlitNumber(numbers, framebuf, crashvx, HSPEED_X, HSPEED_Y);
	if (displayspeed || fabs(crashvy) < SAFE_VY)
	  BlitNumber(numbers, framebuf, crashvy, VSPEED_X, VSPEED_Y);
      }

      /* blit life icons */
      SDL_SetClipRect(framebuf, &lifeiconarea);
      for (loop = 0; loop < livesleft; ++loop)
	BlitSurface(lifeicon, framebuf, LIFEICON_X + 
		    ((lifeicon->w + LIFEICON_SPACING) * loop), LIFEICON_Y);

      /* blit pad */
      SDL_SetClipRect(framebuf, &padarea);
      BlitSurface(padptr, framebuf, padarea.x, padarea.y);

      /* set shadow alpha */
      SDL_SetAlpha(shadow, SDL_SRCALPHA | SDL_RLEACCEL, (MAX_Y - y) * (MAX_SHADOW_ALPHA / MAX_Y));

      /* blit shadow */
      SDL_SetClipRect(framebuf, &shadowarea);
      BlitSurface(shadow, framebuf, pixelx, shadowarea.y);

      /* blit lander */
      SDL_SetClipRect(framebuf, &cliparea);
      BlitSurface(landerfinal, framebuf, pixelx, pixely);

      /* blit fire and set alpha */
      if (crashed) {
	SDL_SetAlpha(fire, SDL_SRCALPHA | SDL_RLEACCEL, firealpha);
	BlitSurface(fire, framebuf, pixelx, pixely);

	firealpha += FIRE_ALPHA * dt;
	if (firealpha > FIRE_ALPHA)
	  firealpha = FIRE_ALPHA;
      }

      /* blit ufo */
      SDL_SetClipRect(framebuf, &ufoarea);
      BlitSurface(ufo, framebuf, ufopixelx, ufopixely);

      /* blit fuelbar */
      if (displayfuel) {
	SDL_SetClipRect(framebuf, &fuelbararea);
	BlitSurface(fuelbar, framebuf, FUELBAR_X, FUELBAR_Y);
      }

      /* swap buffers */
      SDL_Flip(framebuf);
      ++frames;

      /* throttle framerate */
      while ((temptimer = SDL_GetTicks()) - frametimer < 1000.0f / (float)MAX_FPS)
	;

      dt = (temptimer - frametimer) / 1000.0f;
      fps = 1000.0f / (temptimer - frametimer);

      /* check for pause */
      if (CheckForPause(controller))
	frametimer = SDL_GetTicks();
      else
	frametimer = temptimer;

      /* debug stats - uncomment to enable */
      /*if (frames >= floor(fps))
       *fprintf(stderr, "x=%.1f y=%.1f vx=%.1f vy=%.1f ax=%.1f ay=%.1f angle=%.1f mass=%.1f fps=%.1f\n", 
       x, y, vx, vy, ax, ay, angle, mass, fps);*/

      oldpixelx = pixelx;
      oldpixely = pixely;

      /* update joystick */
      SDL_PumpEvents();

      /* update rotation */
      if (! crashed)
	dangle = (dt / ROTATE_PERIOD) * 2.0f * M_PI * GetJoyScale(controller, ROTATION_AXIS);

      angle += dangle;

      if (SDL_JoystickGetButton(controller, THRUST_BUTTON) && mass > START_MASS - FUEL_MASS 
	  && ! onground) {

	/* set pointer to firing sprite and flash lights */

	if (frames >= floor(fps)) {

	  if (spriteptr == lander_ || spriteptr == lander_r)
	    spriteptr = lander_lr;
	  else
	    spriteptr = lander_r;

	  if (padptr == pad_)
	    padptr = pad_l;
	  else
	    padptr = pad_;

	  /* reset framecounter */
	  frames = 0;
	}
	else {

	  if (spriteptr == lander_)
	    spriteptr = lander_r;
	  if (spriteptr == lander_l)
	    spriteptr = lander_lr;
	}

	/* update mass */
	mass -= MAX_THRUST / EXHAUST_V;

	/* mass boundary condition */
	if (mass < START_MASS - FUEL_MASS)
	  mass = START_MASS - FUEL_MASS;

	/* update acceleration */
	ax = (MAX_THRUST * sinf(angle)) / mass;
	ay = (MAX_THRUST * cosf(angle)) / mass;
      }
      else {

	/* set pointer to non-firing sprite and flash lights */

	if (frames >= floor(fps)) {

	  if (spriteptr == lander_ || spriteptr == lander_r)
	    spriteptr = lander_l;
	  else
	    spriteptr = lander_;

	  if (padptr == pad_)
	    padptr = pad_l;
	  else
	    padptr = pad_;

	  /* reset framecounter */
	  frames = 0;
	}
	else {

	  if (spriteptr == lander_r)
	    spriteptr = lander_;
	  if (spriteptr == lander_lr)
	    spriteptr = lander_l;
	}

	ax = ay = 0.0f;
      }

      dvx = ax * dt;
      vx += dvx;
      dx = vx * dt;
      x += dx;

      ay += GRAVITY;
      dvy = ay * dt;
      vy += dvy;
      dy = vy * dt;
      y += dy;

      /* boundary conditions */
      if (y < 0.0f) {
	y = 0.0f;

	onground = 1;

	/* check for a safe landing */
	if (pixelx > (padarea.x - LANDER_XBORDER + PAD_BORDER) 
	    && pixelx < (padarea.x + padptr->w - PAD_BORDER - landerfinal->w + LANDER_XBORDER) 
	    && CheckLanding(vx, vy, angle) && ! crashed) {

	  vx = vy = angle = 0.0f;

	}
	else {

	  /* crash and burn */
	  crashed = 1;

	  if (! gotcrashinfo) {
	    crashvx = vx;
	    crashvy = vy;
	    gotcrashinfo = 1;
	  }

	  dangle = atanf((vx * dt) / LANDER_RADIUS);
	  vx -= vx * CRASH_SLOWDOWN * dt;
	  vy = 0.0f;
	}

      }

      if (y > MAX_Y) {
	y = MAX_Y;

	if (vy > 0.0f)
	  vy = 0.0f;
      }

      if (x > MAX_X)
	x = 0.0f;

      if (x < 0.0f)
	x = MAX_X;

      /* update meteor coordinates */
      for (loop = 0; loop < NUM_METEORS; ++loop) {
	oldmeteorx[loop] = meteorx[loop];
	oldmeteory[loop] = meteory[loop];

	meteorx[loop] -= (SCREEN_HEIGHT / meteorspeed[loop]) * dt;
	meteory[loop] += (SCREEN_HEIGHT / meteorspeed[loop]) * dt;
      }

      /* update meteor timers */
      for (loop = 0; loop < NUM_METEORS; ++loop) {

	if (meteorframes[loop] < 0) {
	  meteorframes[loop] = (MIN_METEOR_SPEED + meteorspeed[loop] + 
				((float)rand() / (float)RAND_MAX)) * fps;

	  meteorx[loop] = (2 * SCREEN_WIDTH * ((float)rand() / (float)RAND_MAX));
	  meteory[loop] = 0 - meteor->h;
	}
	else --meteorframes[loop];
      }

      /* update ufo coordinates */
      oldufopixelx = ufopixelx;
      ufopixelx += ufodirection * (SCREEN_WIDTH / UFO_SPEED) * dt;

      /* update ufo timer */
      if (ufoframes <= 0) {
	ufoframes = (1.0f + UFO_SPEED + (((float)rand() / (float)RAND_MAX)) 
		     * MAX_UFO_DELAY) * fps;

	if (pixelx > SCREEN_WIDTH / 2) {
	  ufodirection = 1;
	  ufopixelx = 0 - ufo->w;
	}
	else {
	  ufodirection = -1;
	  ufopixelx = SCREEN_WIDTH;
	}

      }
      else --ufoframes;

      /* flash fuelbar if low fuel*/
      if (mass - (START_MASS - FUEL_MASS) < FUEL_MASS * FUEL_WARNING) {

	if (fuelbarframes <= 0) {
	  if (displayfuel)
	    displayfuel = 0;
	  else
	    displayfuel = 1;

	  fuelbarframes = fps / FLASH_INTERVAL;
	}
	else --fuelbarframes;
      }

      /* flash speed if crashed */
      if (crashed) {

	if (speedframes <= 0) {
	  if (displayspeed)
	    displayspeed = 0;
	  else
	    displayspeed = 1;

	  speedframes = fps / FLASH_INTERVAL;
	}
	else --speedframes;
      }

      /* delay restart after crash */
      if (crashed) {
	++crashframes;

	if (crashframes > RESTART_DELAY * fps)
	  restartflag = 1;
      }

    }

    --livesleft;
  }

  /* fade out */
  for (loop = 0; loop < FADE_FRAMES; ++loop) {
    SDL_SetAlpha(fadesurface, SDL_SRCALPHA | SDL_RLEACCEL, FADE_ALPHA);
    SDL_SetClipRect(framebuf, NULL);
    BlitSurface(fadesurface, framebuf, 0, 0);
    SDL_Flip(framebuf);
  }

  Mix_CloseAudio();
  QuitSDL();

  exit(0);
}
